package logic;

import java.util.*;

/**
 *  This class is a wrapper for an array of Booleans.
 *  We need a wrapper so that it can be used as a key
 *  in a hashtable. It provides a simple implementation of 
 *  hashCode() for this purpose.
 */
public class BooleanTuple implements BooleanTupleConstants {
	private Boolean [] tuple;
	
	/**
	 *  The constructor accepts a Vector of Conditions -- only
	 *  30 are allowed. It converts the vector to an array and caches it.
	 */
	public BooleanTuple(Vector tuple) throws NestingTooDeepException {
		if(tuple != null && tuple.size() > MAX_LEVELS_OF_NESTING) {
			throw new NestingTooDeepException("Too many conditions to handle safely");
		}
		setTuple(tuple);
	}
	/**
	 *  Two BooleanTuples are equal if they are both nonnull, and their
	 *  underlying instances of Boolean[] have the same length and 
	 *  have entries that have the same booleanValue()'s.
	 */
	public boolean equals(Object ob) {
		if(!(ob instanceof BooleanTuple)) return false;
		BooleanTuple testbt = (BooleanTuple)ob;
		if(ob == null) return false;
		if(tuple == null && testbt.getTuple() == null) return true;
		if(testbt.getTuple() == null) return false;
		if(testbt.size() != size()) return false;
		for(int i = 0; i < size(); ++i) {
			
			Boolean testbtBool = testbt.getTuple()[i];
			Boolean thisBool = tuple[i];
			if(testbtBool.booleanValue() != thisBool.booleanValue()) return false;
		}
		return true;
	}
	public Boolean [] getTuple() {
		return tuple;
	}
	/**
	 *  The algorithm used here is to treat a sequence
	 *  of booleans as a binary sequence, with T=1 and F=0,
	 *  and then treat the binary sequence as an actual
	 *  binary integer. The algorithm simply computes the
	 *  base 10 value of this binary number and returns it.
	 *  An inherent limitation is that the largest allowed int
	 *  is 2^31 -1, so that we can't rely on the algorithm
	 *  for tuples of length > 30.
	 */
	public int hashCode() {
		if(tuple == null) return 0;
		int retVal = 0;
		
		for(int i = 0; i < tuple.length; ++i) {
			Boolean b = tuple[i];
			retVal += num(b) * Math.pow(2,i);
		}
		return retVal;
	}
	private int num(Boolean b) {
		return b.booleanValue() ? 1 : 0;
	}
    /**
     * @author Paul Corazza
     * @param v java.util.Vector
     */
    public void setTuple(Boolean [] arr) {
	    tuple = arr;
    }
    /**
     * @author Paul Corazza
     * @param v java.util.Vector
     */
    public void setTuple(Vector v) {
	    tuple = Util.toBooleanArray(v);
    }
	public int size() {
		return tuple == null ? 0 : tuple.length;
	}
	public String toString() {
	    if(tuple == null) return "null";
	    StringBuffer accum = new StringBuffer();
	    for(int i = 0; i < tuple.length; ++i) {
	        int num = num(tuple[i]);
	        accum.append( ((num == 1) ? "T" : "F") );
	    }
	    return accum.toString();
	}
}