package logic;

import java.util.*;
import java.io.*;

public class BooleanTupleFactory {
	
	/** 
	 *  This method generates the array of BooleanTuples that is generated from
	 *  the boolean String tf (like "TT**F") -- a String that may contain '*'s.
	 *  We first generate all possible boolean strings that don't have stars,
	 *  then convert to BooleanTuples. 
	 *  <p>
	 *  For example, if tf is "TT**F", this method first generates an array
	 *  of four non-star boolean Strings as follows:
	 *  <pre>
	 *     0:  "TTTTF"
	 *     1:  "TTTFF"
	 *     2:  "TTFTF"
	 *     3:  "TTFFF"
	 *  </pre>
	 *  These four rows are thought of as a 2-d array; the 0th "column" in this
	 *  example would be
	 *  <pre>
	 *     T
	 *     T
	 *     T
	 *     T
	 *  </pre>
	 *  In order to fill in the '*'s with all possible T's and F's, we use the
	 *  truthTable(..) method.
	 *  NOTE: Error handling is not being done in case conversion fails for some reason.
	 */
	public static BooleanTuple [] getTuples(String tf) throws NestingTooDeepException {
		
		if(tf == null || tf.equals("")) return null;
		if(tf.length() > BooleanTupleConstants.MAX_LEVELS_OF_NESTING){
			throw new NestingTooDeepException("Can't have more than "+
				   BooleanTupleConstants.MAX_LEVELS_OF_NESTING + " levels of nesting");    
		}
		int numTuples = countTuples(tf);
		int len = tf.length();
		int numStars = countStars(tf);
		int [] indicesOfStarredCols = {};
		if(numStars > 0) {
		    indicesOfStarredCols = computeIndicesOfStarredCols(tf);
		}

		StringBuffer [] tfArr = new StringBuffer[numTuples];
		BooleanTuple [] btArr = new BooleanTuple[numTuples];
		
		for(int col = 0; col < len; ++col) {
			for(int row = 0; row < numTuples; ++row) {
				if(col == 0) tfArr[row] = new StringBuffer("");				
				char c = tf.charAt(col);
				
				tfArr[row].append(obtainBoolChar(col, row, c, indicesOfStarredCols));
			}
		}
		for(int row = 0; row < numTuples; ++row) {
			btArr[row] = getTuple(tfArr[row].toString());
		}
		return btArr;
	}
	
    /**
     *  This function computes the truth value (true = 0, false = 1) that
     *  appears in a standard truth table having numStarCols propositional symbols,
     *  where starCol is the requested column index and row is the requested row 
     *  index. For instance, for three prop symbols, the truth table looks like this:
     *  <pre>
     *      T   T   T
     *      F   T   T
     *      T   F   T
     *      F   F   T
     *      T   T   F
     *      F   T   F
     *      T   F   F
     *      F   F   F
     *  </pre>
     *  The way the algorithm works is this: To compute values in a truth table
     *  having n columns, the algorithm makes use of the fact that to obtain
     *  the values in columns 0...n-2, you can use computations for the case of
     *  n-1 columns. This just leaves the n-1st column to be computed. But (as you
     *  can see from the example) the n-1st column is easy to compute -- return value
     *  0 if the requested row is in the first half of the rows (i.e., less than 
     *  2^(total_num_columns-1)), or value 1 if it's in the second half.
     *  <p>
     *  @param starCol - the index of the requested column
     *  @param row - the index of the requested tow
     *  @param numStarCols - total number of columns, which is the total num of prop symbols
     *  @return int - either 0 (true) or 1 (false)
     */
    public static int truthTable(int starCol, int row, int numStarCols) {
        //validate input 
        if(starCol < 0 || starCol > numStarCols - 1 || row < 0 || 
            row > Math.pow(2,numStarCols)-1 || numStarCols < 1) {
                
                System.out.println("Invalid input ["+starCol+", "+row+", "+numStarCols+"] for truthTable method");
                return -1;
        }
        //if starCol is largest possible, return 0 if row lies in first half of the rows, return
        //1 if in second half
        if( starCol == numStarCols - 1) {
            return (row < Math.pow(2,numStarCols-1)) ?
                   0 :
                   1;
        }
        //if starCol is one of the smaller possible values, use the version of truthTable
        //for which total number of columns is 1 less than the current value
        else {
            return truthTable(starCol, row % ((int)Math.pow(2,numStarCols-1)), numStarCols-1);
        }
    }
    
    //------------ PRIVATE METHODS --------------//
    
	/**
	 *  If character c is just 'T' or 'F', this method just returns c. If c is a star, however,
	 *  then this method converts the column index of this star to the index of the
	 *  array indicesOfStarredCols that contains it, and then calls truthTable(..) to 
	 *  determine whether, for the  given column and row, the value should be 'T' or 'F'
	 *  NOTE: No error handling is done in case incorrect data is being passed around.
	 */
	private static char obtainBoolChar(int col, int row, char c, int[] indicesOfStarredCols) {
		switch(c) {
			case 'T':
			case 't':
				return 'T';

			case 'F':
			case 'f':
				return 'F';

			case '*':
			    //find index of starred col index in the indicesOfStarredCols array
			    int adjustedColNum = Util.findIndex(col, indicesOfStarredCols);
			    
			    //call truthTable to determine appropriate truth value for this row and column
			    int booleanAsNumeric = 
			        truthTable(adjustedColNum, row, indicesOfStarredCols.length);
			        
			    //convert to 'T' or 'F'
				return (booleanAsNumeric == 0) ? 'T' : 'F';

			default:
				break;
		}
		return 'F';
	}
    
    
    /**
     *  Counts the number of '*'s that occur in s
     */
	private static int countStars(String s) {
		int count = 0;
		if(s == null) return count;
		for(int i = 0; i < s.length(); ++i) {
			if(s.charAt(i) == '*') ++count;
		}
		return count;
	}
	
	/**
	 *  Places the positions of '*' in s into an array. Returns an empty (non-null)
	 *  array if there are no '*'s.
	 */
    private static int [] computeIndicesOfStarredCols(String s) {
        int arraySize = countStars(s);
        int [] arrayOfIndices = new int[arraySize];
        int currIndex = 0;
        for(int i = 0; i < s.length(); ++i) {
            if(s.charAt(i) == '*') {
                arrayOfIndices[currIndex] = i;
                ++currIndex;
            }
        }
        return arrayOfIndices;
    }
	
	/**
	 *  Counts the number of boolean strings (like "TFFTF") that will be derived
	 *  from s. If s has no stars, the return value is 1. If 1 star, the value is
	 *  2. In general, the return value is 2^num_stars
	 */
	private static int countTuples(String s) {
		return (int)Math.pow(2,countStars(s));
	}
	
	/**
	 *  Converts 'T' and 'F' to the corresponding Booleans
	 */
	private static Boolean getBoolean(char c) {
		if(c == 't' || c == 'T') return Boolean.TRUE;
		return Boolean.FALSE;
	}
	/**
	 *  This method returns a vector of Booleans, wrapped in a BooleanTuple, that
	 *  corresponds to a boolean String like "TTFFF". Used by getTuples(..).
	 *  This method does not handle Strings that contain stars, and assumes
	 *  that all characters in tf are either 'T' or 'F'
	 */
	private static BooleanTuple getTuple(String tf) throws NestingTooDeepException {
		
		if(tf == null || tf.equals("")) return null;
		if(tf.length() > BooleanTupleConstants.MAX_LEVELS_OF_NESTING){
			throw new NestingTooDeepException("Can't have more than "+
				   BooleanTupleConstants.MAX_LEVELS_OF_NESTING + " levels of nesting");    
		}
		int len = tf.length();
		Vector v = new Vector(len);
		
		for(int i = 0; i < len; ++i) {
			char c = tf.charAt(i);
			Boolean b = getBoolean(c);
			v.addElement(b);
		}
		return new BooleanTuple(v);
	}
    
    /**
     *  The purpose of this main method is to verify that the truthTable(..) method
     *  produces correct output. In body of main, we attempt to output the entire
     *  truth table for 3 propositional symbols (with true = 0, false = 1).
     *  Output appears in a file c:\out.txt.
     */
    public static void main(String [] args) {
        try {
            FileWriter fw = new FileWriter(new File("c:\\out.txt"));
            PrintWriter out = new PrintWriter(fw);
            out.println("This is the truth table for 4 prop symbols");
            for(int r = 0; r < 16; ++r) {
                for(int c = 0; c < 4; ++c) {
                    out.print("  " + truthTable(c,r,4));
                }
                out.println("");
            }
            out.close();
        }
        catch(IOException ie) {}
    }        
	
}